/*
 * $Id: //depot/prod/zephyr/lufthansaair/src/java/netapp/manage/NaSSLException.java#1 $
 *
 * Copyright (c) 2011 NetApp, Inc.
 * All rights reserved.
 *
 * ONTAPI Server stuff
 */

package netapp.manage;

import javax.net.ssl.SSLHandshakeException;
import java.security.cert.X509Certificate;

/**
 *  Indicates some kind of error detected during SSL handshake. 
 *  This exception will be thrown when there is a problem in 
 *  verifying the remote certificate chain.
 */

public class NaSSLException extends SSLHandshakeException {

	private X509Certificate[] _remoteCertificate;

	/**
	 *  Constructs a new <code>NaSSLException</code> with
	 *  the specified remote certificate chain and the 
	 *  reason message for verification failure.
	 *
	 *  @param	chain		the remote certificate chain.
	 *  @param	reason		the reason message.
	 */
	public NaSSLException(X509Certificate[] chain, String reason)
	{
		super(reason);
		_remoteCertificate = chain;
	}

	/**
	 *  Returns the chain of certificates that is sent by the peer.
	 *
	 *  @return	The certificate chain.
	 */
	public X509Certificate[] getCertificateChain()
	{
		return _remoteCertificate;
	}
}
