/*
 * $Id: //depot/prod/zephyr/lufthansaair/src/java/netapp/manage/http/HTTPClient.java#1 $
 *
 * Copyright (c) 2001-2004 Network Appliance, Inc.
 * All rights reserved.
 *
 * ONTAPI Server stuff
 */

package netapp.manage.http;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.net.BindException;
import java.net.Socket;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

import javax.net.SocketFactory;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLContext;
import javax.net.ssl.X509TrustManager;

import java.security.InvalidAlgorithmParameterException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.cert.CertPath;
import java.security.cert.CertPathValidator;
import java.security.cert.CertPathValidatorException;
import java.security.cert.CertificateFactory;
import java.security.cert.PKIXParameters;
import java.security.cert.X509Certificate;
import java.security.cert.CertificateException;
import java.net.SocketAddress;
import java.net.InetSocketAddress;
import java.net.SocketTimeoutException;
import java.lang.IllegalArgumentException;
import java.nio.channels.IllegalBlockingModeException;

import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.KeyManager;

import java.security.KeyStore;

import javax.net.ssl.TrustManagerFactory;
import javax.net.ssl.TrustManager;

import sun.security.util.HostnameChecker;
import netapp.manage.NaCertificateException;

import java.io.FileInputStream;
import java.security.GeneralSecurityException;

/**
 *  An HTTP client implements the client portion of RFC2616.
 */
public class HTTPClient {

	/**
	 *  Custom trust manager class used to authenticate the remote side of a secure socket.
	 */
	static class CustomX509TrustManager implements X509TrustManager
	{
		private X509TrustManager _trustManager;
		private String _hostNameOrIP;
		private boolean _isHostnameVerificationEnabled;
		private KeyStore _truststore;

		public CustomX509TrustManager(AuthInfo authInfo) throws 
			GeneralSecurityException, IOException
		{
			_trustManager = null;
			FileInputStream is = null;
			_truststore = null;
			_hostNameOrIP = authInfo.getHost();
			_isHostnameVerificationEnabled = authInfo.isHostnameVerificationEnabled();
			try
			{
				TrustManagerFactory tmf = TrustManagerFactory.getInstance(
					TrustManagerFactory.getDefaultAlgorithm());

				if (authInfo.getTrustStoreFile() != null) {
					is = new FileInputStream(authInfo.getTrustStoreFile());
					KeyStore tks = KeyStore.getInstance(authInfo.getTrustStoreType());    
					tks.load(is, null);
					is.close();
					tmf.init(tks);
					_truststore = tks;
				} else {
					tmf.init((KeyStore) null);
				}
				TrustManager tms [] = tmf.getTrustManagers();
				for (int i = 0; i < tms.length; i++) {
					if (tms[i] instanceof X509TrustManager) {
						_trustManager = (X509TrustManager) tms[i];
						break;
					}
				}
			} catch (GeneralSecurityException e) {
				throw e;
			} catch (IOException e) {
				throw e;
			} finally {
				if (is != null)
					is.close();
			}
		}

		public void checkClientTrusted(X509Certificate[] chain, String authType)
									throws CertificateException 
		{
			_trustManager.checkClientTrusted(chain, authType);
		}

		public void checkServerTrusted(X509Certificate[] chain, String authType)
									throws CertificateException 
		{
			X509Certificate cert = (X509Certificate)chain[0];
			try 
			{
				_trustManager.checkServerTrusted(chain, authType);
				cert.checkValidity();

				// Validate the incoming certificate chain that is sent 
				// by the server against the list of trusted certificates 
				// in the trust store.

				// This needs to be done only for Signed certificates
				if (cert.getSubjectDN().equals(cert.getIssuerDN()) == false) {
					if (_truststore == null) {
						File file = null;
						FileInputStream is = null;
						String filename = System.getProperty("javax.net.ssl.trustStore");
						
						if (filename != null) { 
							is = new FileInputStream(filename); 
						} else {
							filename = System.getProperty("java.home") + 
									"/lib/security/jssecacerts".replace('/', File.separatorChar);
							file = new File(filename);
							if (!file.exists()) {
								filename = System.getProperty("java.home") + 
										"/lib/security/cacerts".replace('/', File.separatorChar);
								file = new File(filename);
							}
							is = new FileInputStream(file);
						}
						KeyStore keystore = KeyStore.getInstance(KeyStore.getDefaultType());
						keystore.load(is, null);
						is.close();
						_truststore = keystore;
					}
					// Populate the trusted CAs from the trusted certificate entries 
					// contained in the specified truststore. 
					PKIXParameters params = new PKIXParameters(_truststore);
					params.setRevocationEnabled(false);
					CertificateFactory certFactory = CertificateFactory.getInstance("X.509");
					// Build the CertPath object from a list of remote certificates
					CertPath certPath = certFactory.generateCertPath(Arrays.asList(chain));
					CertPathValidator certPathValidator = CertPathValidator.getInstance(CertPathValidator.getDefaultType());
					// Validate the specified certification path using the specified 
					// parameter set.
					certPathValidator.validate(certPath, params);
				}
				if (_isHostnameVerificationEnabled) {
					HostnameChecker.getInstance(HostnameChecker.TYPE_TLS).match(_hostNameOrIP, cert);
				}
			} catch (FileNotFoundException e) {
				throw new NaCertificateException(chain, e.getMessage());
			} catch (KeyStoreException e) {
				throw new NaCertificateException(chain, e.getMessage());
			} catch (NoSuchAlgorithmException e) {
				throw new NaCertificateException(chain, e.getMessage());
			} catch (IOException e) {
				throw new NaCertificateException(chain, e.getMessage());
			} catch (InvalidAlgorithmParameterException e) {
				throw new NaCertificateException(chain, e.getMessage());
			} catch (CertPathValidatorException e) {
				throw new NaCertificateException(chain, e.getMessage());
			} catch (CertificateException e) {
				String msg = e.getMessage();
				// Currently DFM certificates doesn't contain subject alt names field.
				// Throw a better error message when there is no such field.
				if (msg != null && msg.equals("No subject alternative names present")) {
					String subjPrincipal = cert.getSubjectX500Principal().getName();
					String name = "";
					int index1, index2;
					StringTokenizer st = new StringTokenizer(subjPrincipal);
					while (st.hasMoreTokens()) {
						String tok = st.nextToken();
						index1 = tok.indexOf("CN=");
							if (index1 >= 0) {
								index2 = tok.indexOf(',', index1 + 3);
								if(index2 == -1) {
									index2 = tok.length();
								}
								name = tok.substring(index1 + 3, index2);
								break;
							}
					}
					// Return success if the IPAddress of the server matches 
					// with that of Common Name in the certificate.
					if (_hostNameOrIP.equalsIgnoreCase(name)) {
						return;
					}
					msg = "Server certificate name (CN=" + name + 
					"), hostname (" + _hostNameOrIP + ") mismatch.";
					throw new NaCertificateException(chain, msg);
				} else {
					throw new NaCertificateException(chain, msg);
				}
			}
		}

		public X509Certificate[] getAcceptedIssuers()
		{ 
			return _trustManager.getAcceptedIssuers();
		}
	}

	private static final long TIMEOUT = 30000; /* milliseconds */

	private static Object _lock;
	private static SocketFactory _plainFactory;
	private SocketFactory _sslFactory;
	private static SocketFactory _defaultSSLFactory;
	private boolean _useTLS = false;
	private boolean _useSSL = false;

	/**
	 *  The debug level. Set this to a positive integer and
	 *  information about HTTP transactions will be printed
	 *  to {@link java.lang.System#err}.
	 */
	public static int snoop;

	static
	{
		_lock = new Object();
		_plainFactory = SocketFactory.getDefault();
		_defaultSSLFactory = null; /* delay the creation of the SSL factory */
		snoop = 0;
	}

	/**
	 *  Creates a SSL socket factory. The socket factory creates SSL
	 *  sockets with trust managers that can either verify the remote 
	 *  certificate or can blindly accepts any certificate supplied by the server.
	 *  @param	authInfo	Information needed for CBA and for server 
	 *                      certificate verification.
	 *  @return	A socket factory, or <code>null</code> if no suitable 
	 *  SSL provider exists.
	 */
	private static SocketFactory makeSSLFactory(AuthInfo authInfo) throws 
					GeneralSecurityException, IOException
	{
		SocketFactory sslFactory = null;
		KeyManager[] km = null;
		TrustManager[] tm = null;
		boolean useDefaultSSLFactory = true;

		try
		{
			// The default socket factory is used when we donot need 
			// certificate based authentication or server certificate verification.
			if (authInfo != null && 
				(authInfo.isCBAEnabled() || authInfo.isServerCertVerificationEnabled())) {
				useDefaultSSLFactory = false;
			}
			if (useDefaultSSLFactory && _defaultSSLFactory != null) {
				return _defaultSSLFactory;
			}
			if (authInfo != null && authInfo.isCBAEnabled() && 
					authInfo.getKeyStoreFile() != null) {
				km = getKeyManagers(authInfo);
			}
			if (authInfo != null && authInfo.isServerCertVerificationEnabled()) {
				tm = getTrustManagers(authInfo);
			}
			else {
				tm = getDefaultTrustManagers();
			}
			SSLContext ctx = SSLContext.getInstance("SSL");
			ctx.init(km, tm, new SecureRandom());
			sslFactory = ctx.getSocketFactory();
			if (useDefaultSSLFactory) {
				_defaultSSLFactory = sslFactory;
			}
			return sslFactory;
		} catch (GeneralSecurityException e) {
			throw e;
		} catch (IOException e) {
			throw e;
		}
	}

	/**
	 *  Creates trust managers that blindly accepts any certificate supplied 
	 *  by the server.
	 *  @return	 List of X509TrustManagers.
	 */
	private static X509TrustManager[] getDefaultTrustManagers()
	{
		X509TrustManager[] tm = { new X509TrustManager() {
		public void checkClientTrusted(X509Certificate[] chain, String authType)
			throws CertificateException {}
		public void checkServerTrusted(X509Certificate[] chain, String authType)
			throws CertificateException {}
		public X509Certificate[] getAcceptedIssuers()
			{ return new X509Certificate[0]; } }
		};
		return tm;
	}

	/**
	 *  Creates trust managers that are required to verify the certificate 
	 *  supplied by the remote server.
	 * @param	authInfo	Information needed for CBA and for server 
	 *						certificate verification.
	 *  @return	 List of X509TrustManagers.
	 */
	private static X509TrustManager[] getTrustManagers(AuthInfo authInfo) 
				throws GeneralSecurityException, IOException
	{
		X509TrustManager[] tm = new X509TrustManager[] {
					new CustomX509TrustManager(authInfo)};
		return tm;
	}

	/**
	 *  Creates key managers where the client keys and certificates are kept.
	 *  Both the keystore and individual private keys should be password protected.
	 * @param	authInfo Information needed for CBA and for server 
	 *                 certificate verification.
	 *  @return	List of X509TrustManagers.
	 */
	private static KeyManager[] getKeyManagers(AuthInfo authInfo) throws 
			GeneralSecurityException, IOException
	{
		KeyManager[] km = null;
		FileInputStream is = null;
		KeyStore ks = null;

		try
		{
			KeyManagerFactory kmf = 
			KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
			is = new FileInputStream(authInfo.getKeyStoreFile());
			ks = KeyStore.getInstance(authInfo.getKeyStoreType());
			ks.load(is, authInfo.getKeyStorePasswd() != null ? 
					authInfo.getKeyStorePasswd().toCharArray() : null);
			is.close();
			kmf.init(ks, authInfo.getKeyPasswd() != null ? 
					authInfo.getKeyPasswd().toCharArray() : null);
			km = kmf.getKeyManagers();
		} catch(IOException e) {
			throw e;
		} catch(GeneralSecurityException e) {
			throw e;
		} finally {
			if (is != null)
				is.close();
		}
		return km;
	}

	private boolean _ssl;
	private InetAddress _host;
	private int _port;

	private Socket _sock;
	private OutputStream _out;
	private InputStream _in;

	private int _useCount;
	private long _lastTimeUsed;
	private int _timeOut;
	private AuthInfo _authInfo;

	/**
	 *  Constructs a <code>HTTPClient</code>. A single socket
	 *  connection is made to the HTTP server,
	 *
	 *  @param	ssl		flag to use HTTPS instead of HTTP.
	 *  @param	host	the server hostname.
	 *  @param	port	the server TCP port.
	 *  @exception	IOException	if the connection cannot be established.
	 *  @exception	UnknownHostException	if the hostname is not known.
	 */
	public HTTPClient(boolean ssl, String host, int port)
		throws IOException, UnknownHostException
	{
		this(ssl, host, port, false);
	}

	/**
	 *  Constructs a <code>HTTPClient</code>. A single socket
	 *  connection is made to the HTTP server,
	 *
	 *  @param	ssl		flag to use HTTPS instead of HTTP.
	 *  @param	host	the server hostname.
	 *  @param	port	the server TCP port.
	 *  @param	reservedPort	flag indicating to bind to a reserved port.
	 *  @exception	IOException	if the connection cannot be established.
	 *  @exception	UnknownHostException	if the hostname is not known.
	 */
	public HTTPClient(boolean ssl, String host, int port,
		boolean reservedPort) throws IOException, UnknownHostException
	{
		this(ssl, InetAddress.getByName(host), port, reservedPort, 0, null, false);
	}

	/**
	 *  Constructs a <code>HTTPClient</code>. A single socket
	 *  connection is made to the HTTP server,
	 *
	 *  @param	ssl		flag to use HTTPS instead of HTTP.
	 *  @param	host	the server address.
	 *  @param	port	the server TCP port.
	 *  @exception	IOException	if the connection cannot be established.
	 */
	public HTTPClient(boolean ssl, InetAddress host, int port)
		throws IOException
	{
		this(ssl, host, port, false, 0, null, false);
	}
    
    /**
	 *  Constructs a <code>HTTPClient</code>. A single socket
	 *  connection is made to the HTTP server,
	 *
	 *  @param	ssl		flag to use HTTPS instead of HTTP.
	 *  @param	host	the server address.
	 *  @param	port	the server TCP port.
	 *  @param	reservedPort	flag indicating to bind to a reserved port.
	 *  @exception	IOException	if the connection cannot be established.
	 */
 	public HTTPClient(boolean ssl, InetAddress host, int port,
		boolean reservedPort) throws IOException
    {
		this(ssl, host, port, reservedPort, 0, null, false);
	}

	/**
	 *  Constructs a <code>HTTPClient</code>. A single socket
	 *  connection is made to the HTTP server,
	 *
	 *  @param	ssl		flag to use HTTPS instead of HTTP.
	 *  @param	host	the server address.
	 *  @param	port	the server TCP port.
	 *  @param	timeOut	the connection timeout value.
	 *  @param	reservedPort	flag indicating to bind to a reserved port.
	 *  @exception	IOException	if the connection cannot be established.
	 *  @exception	SocketTimeoutException	if timeout expires before 
	 *  					connecting.
	 *  @exception	IllegalArgumentException	if endpoint is null or 
	 *  					is a SocketAddress subclass 
	 *  					not supported by this socket. 
	 *  @exception	IllegalBlockingModeException	if this socket has an 
					associated channel, and the channel is 
					in non-blocking mode.
	 */
	public HTTPClient(boolean ssl, InetAddress host, int port,
		boolean reservedPort,int timeOut) throws IOException, 
	SocketTimeoutException, IllegalArgumentException, 
	IllegalBlockingModeException
	{
		this(ssl, host, port, reservedPort, timeOut, null, false);
	}

	/**
	 *  Constructs a <code>HTTPClient</code>. A single socket
	 *  connection is made to the HTTP server,
	 *
	 *  @param	ssl		flag to use HTTPS instead of HTTP.
	 *  @param	host	the server address.
	 *  @param	port	the server TCP port.
	 *  @param	reservedPort	flag indicating to bind to a reserved port.
	 *  @param	timeOut	the connection timeout value.
	 *  @param	authInfo	certificate bsed authentication info structure.
	 *  @exception	IOException	if the connection cannot be established.
	 *  @exception	SocketTimeoutException	if timeout expires before 
	 *  					connecting.
	 *  @exception	IllegalArgumentException	if endpoint is null or 
	 *  					is a SocketAddress subclass 
	 *  					not supported by this socket. 
	 *  @exception	IllegalBlockingModeException	if this socket has an 
					associated channel, and the channel is 
					in non-blocking mode.
	 */
	public HTTPClient(boolean ssl, InetAddress host, int port,
		boolean reservedPort,int timeOut, AuthInfo authInfo) throws IOException, 
			SocketTimeoutException, IllegalArgumentException, 
			IllegalBlockingModeException
	{
		this(ssl, host, port, reservedPort, timeOut, authInfo, false);
	}

	/**
	 *  Constructs a <code>HTTPClient</code>. A single socket
	 *  connection is made to the HTTP server,
	 *
	 *  @param	ssl		flag to use HTTPS instead of HTTP.
	 *  @param	host	the server address.
	 *  @param	port	the server TCP port.
	 *  @param	reservedPort	flag indicating to bind to a reserved port.
	 *  @param	timeOut	the connection timeout value.
	 *  @param	authInfo	certificate bsed authentication info structure.
	 *  @param	useTLS	flag to enable TLS protocol during SSL communication.
	 *  @exception	IOException	if the connection cannot be established.
	 *  @exception	SocketTimeoutException	if timeout expires before
	 *  					connecting.
	 *  @exception	IllegalArgumentException	if endpoint is null or
	 *  					is a SocketAddress subclass
	 *  					not supported by this socket.
	 *  @exception	IllegalBlockingModeException	if this socket has an
					associated channel, and the channel is
					in non-blocking mode.
	 */
	public HTTPClient(boolean ssl, InetAddress host, int port,
			boolean reservedPort,int timeOut, AuthInfo authInfo, boolean useTLS) throws IOException, 
				SocketTimeoutException, IllegalArgumentException, 
				IllegalBlockingModeException
	{
		this(ssl, host, port, reservedPort, timeOut, authInfo, useTLS,false);
	}

	/**
	 *  Constructs a <code>HTTPClient</code>. A single socket
	 *  connection is made to the HTTP server,
	 *
	 *  @param	ssl		flag to use HTTPS instead of HTTP.
	 *  @param	host	the server address.
	 *  @param	port	the server TCP port.
	 *  @param	reservedPort	flag indicating to bind to a reserved port.
	 *  @param	timeOut	the connection timeout value.
	 *  @param	authInfo	certificate bsed authentication info structure.
	 *  @param	useTLS	flag to enable TLS protocol during SSL communication.
	 *  @param	useSSL	flag to enable SSLv3 protocol during SSL communication.
	 *  @exception	IOException	if the connection cannot be established.
	 *  @exception	SocketTimeoutException	if timeout expires before
	 *  					connecting.
	 *  @exception	IllegalArgumentException	if endpoint is null or
	 *  					is a SocketAddress subclass
	 *  					not supported by this socket.
	 *  @exception	IllegalBlockingModeException	if this socket has an
					associated channel, and the channel is
					in non-blocking mode.
	 */
	
	public HTTPClient(boolean ssl, InetAddress host, int port,
			boolean reservedPort,int timeOut, AuthInfo authInfo, boolean useTLS,boolean useSSL) throws IOException, 
				SocketTimeoutException, IllegalArgumentException, 
				IllegalBlockingModeException
	{
		_useTLS = useTLS;
		_useSSL = useSSL;
		_ssl = ssl;
		_host = host;
		_port = port;
		// convert timeout value in milli-seconds
		_timeOut = timeOut * 1000;
		_authInfo = authInfo;
		
		if (_ssl) {
			synchronized (_lock) {
				/*
				 *  We create the SSL factory just in time, only
				 *  when it's needed.
				 */
				try
				{
					_sslFactory = makeSSLFactory(authInfo);
				}catch (GeneralSecurityException e) {
					throw (IOException) new IOException(e.toString()).initCause(e);
				} catch (IOException e) {
					throw e;
				}
			}
			if(_timeOut > 0) {
				_sock = _sslFactory.createSocket();
				SocketAddress sockAddress = 
					new InetSocketAddress(_host, _port);
				_sock.connect(sockAddress, _timeOut);
				_sock.setSoTimeout(_timeOut);
			}
			else {
				_sock = _sslFactory.createSocket(_host, _port);
			}
		} else if (reservedPort) {
			for (int localPort = 1023; null == _sock; localPort--) {
				if (0 == localPort) {
					throw new IOException(
						"Failed to bind to a reserved port");
				}
				try {
					if(_timeOut > 0) {
						_sock = _plainFactory.createSocket();
						SocketAddress sockAddress = 
							new InetSocketAddress(_host, _port);
						InetAddress addr = null;
						SocketAddress bindSockAddress = 
							new InetSocketAddress(addr, localPort);
						_sock.bind(bindSockAddress);
						_sock.connect(sockAddress, _timeOut);
						_sock.setSoTimeout(_timeOut);
					}
					else {
						_sock = _plainFactory.createSocket(_host, _port,
						null, localPort);
					}                
				} catch (BindException be) {
					/*
					 * the local port is in use, try other ports
					 */
					_sock = null;
				}
			}
		} else {
            if(_timeOut > 0) {
				_sock =  _plainFactory.createSocket();
				SocketAddress sockAddress = 
					new InetSocketAddress(_host, _port);
                _sock.connect(sockAddress, _timeOut);
				_sock.setSoTimeout(_timeOut);
            }
            else {
                _sock = _plainFactory.createSocket(_host, _port);
            }  
		}

		/*
		 *  burt147328: don't include TLSv1 in the list of enabled
		 *  protocols, because the SSL handshake barfs when talking
		 *  to filers with Ontap 7.0.
		 *  (This is really a workaround for burt146999.)
		 */
		if (_sock instanceof SSLSocket) {
			SSLSocket sslsock = (SSLSocket) _sock;

			String[] protocols = sslsock.getEnabledProtocols();
			List protocolList = new ArrayList(Arrays.asList(protocols));
			if(!_useSSL) {
				protocolList.remove("SSLv3");
			}
			
			if (!_useTLS) {
				Iterator protocolIterator = protocolList.iterator();
				while (protocolIterator.hasNext()) {
					String protocol = (String)protocolIterator.next();
					if(protocol.indexOf("TLS") != -1) {
						protocolIterator.remove();
					}
				}
			}
			

			/**
			 *  Remove SSLv2 protocol to enforce stronger communication.
			 */
			protocolList.remove("SSLv2Hello");
			protocols = (String[]) protocolList.toArray(new String[0]);
			sslsock.setEnabledProtocols(protocols);

			List cipherList = new ArrayList();
			String[] ciphers = sslsock.getEnabledCipherSuites();
			for(int i = 0; i < ciphers.length; i++) {
				/**
				 *  Removing MD5 enabled ciphers due to
				 *  known vulnerability (CVE-2004-2761).
				 */
				if (ciphers[i].indexOf("MD5") == -1) {
					cipherList.add(ciphers[i]);
				}
			}
			ciphers = (String[]) cipherList.toArray(new String[0]);
			sslsock.setEnabledCipherSuites(ciphers);
		}

		_out = _sock.getOutputStream();
		_in = _sock.getInputStream();
		_useCount = 0;
		_lastTimeUsed = System.currentTimeMillis();
	}

	
	/**
	 *  Close this HTTP client's connections. Call this function to free
	 *  resources after you are done using the <code>HTTPClient</code>.
	 */
	public void close()
	{
		try {
			_out.close();
		} catch (IOException e) {
			/* XXX - log this somewhere useful */
			e.printStackTrace();
		}
		try {
			_in.close();
		} catch (IOException e) {
			/* XXX - log this somewhere useful */
			e.printStackTrace();
		}
		try {
			_sock.close();
		} catch (IOException e) {
			/* XXX - log this somewhere useful */
			e.printStackTrace();
		}
	}

	/**
	 *  Test whether this <code>HTTPClient</code> was created with some
	 *  particular paramters.
	 *
	 *  @param	ssl		flag to use HTTPS instead of HTTP.
	 *  @param	host		the server hostname.
	 *  @param	port		the server TCP port.
	 *  @param	authInfo	certificate bsed authentication info structure.
	 *  @return	<code>true</code> if all the parameters are the
	 *		same as when the object was created, otherwise
	 *		<code>false</code>.
	 */
	public boolean isCompatible(boolean ssl, String host, int port, AuthInfo authInfo)
	{
		return ((ssl == _ssl) && host.equals(_host.getHostName()) && 
				(port == _port) && AuthInfo.compare(_authInfo, authInfo));
	}

	/**
	 *  Test whether this <code>HTTPClient</code> was created with some
	 *  particular paramters.
	 *
	 *  @param	ssl		flag to use HTTPS instead of HTTP.
	 *  @param	host		the server address.
	 *  @param	port		the server TCP port.
	 *  @param	authInfo	certificate bsed authentication info structure.
	 *  @param	useTLS	flag to enable TLS protocol during SSL communication.
	 *  @return	<code>true</code> if all the parameters are the
	 *		same as when the object was created, otherwise
	 *		<code>false</code>.
	 */
	public boolean isCompatible(boolean ssl, InetAddress host, int port, AuthInfo authInfo, boolean useTLS)
	{
		return ((ssl == _ssl) && host.equals(_host) && 
				(port == _port) && AuthInfo.compare(_authInfo, authInfo) && 
				(useTLS == _useTLS));
	}

	/**
	 *  Test whether the connection has been idle for too long. A
	 *  connection that has timed out should be closed and discarded.
	 *
	 *  @return	<code>true</code> if the connection was idle for
	 *		too long, otherwise <code>false</code>.
	 */
	public boolean isTimedOut()
	{
		return System.currentTimeMillis() > (_lastTimeUsed + TIMEOUT);
	}

	/**
	 *  Get the number of requests that have been processed by this
	 *  <code>HTTPClient</code>.
	 *
	 *  @return	number of requests.
	 */
	public int getUseCount()
	{
		return _useCount;
	}

	/**
	 *  Enables TLS protocol for use on connection over HTTPS Transport.
	 *  By default, TLS protocol is enabled.
	 */
	public void enableTLS()
	{
		_useTLS = true;
	}

	/**
	 *  Disables TLS protocol for use over HTTPS Transport.
	 *  By default, TLS protocol is enabled.
	 */
	public void disableTLS()
	{
		_useTLS = false;
	}

	/**
	 *  Determines whether TLS protocol is enabled for use over HTTPS Transport.
	 */
	public boolean isTLSEnabled()
	{
		return _useTLS;
	}

	
	/**
	 *  Send a <code>HTTPRequest</code> and receive the corresponding
	 *  <code>HTTPResponse</code>. This function will automatically
	 *  set the "Host" header in the request based on the parameters
	 *  of this <code>HTTPClient</code>.
	 *
	 *  @param	request		the request to send.
	 *  @return	the response recieved.
	 *  @exception	IOException	if there is a communication error.
	 *  @exception	HTTPException	if there is a HTTP protocol error.
	 */
	public HTTPResponse doRequest(HTTPRequest request)
		throws IOException, HTTPException
	{
		request.setHeader("Host", _host.getHostName() +
			((80 == _port) ? "" : (":" + _port)));
		request.write(_out);
		HTTPResponse response = HTTPResponse.read(_in);
		_useCount++;
		_lastTimeUsed = System.currentTimeMillis();
		return response;
	}
}
